#include "event.h"
#include "log.h"
#include "signal.h"
#include "uhid.h"
#include "uinput.h"

#include <jni.h>

#include <errno.h>
#include <poll.h>
#include <pthread.h>
#include <stdbool.h>
#include <time.h>

// Whether the user runs as root:
// if true, use uinput which is simpler;
// otherwise, attempt to work with uhid which usually requires lower permissions
#define IS_ROOT_USER false

// TODO: Determine the method automatically by trying to create a uinput device
//  and avoid calls to this library if it is impossible to create a uhid device either
#if IS_ROOT_USER
#define USER_SPACE_METHOD uinput
#else
#define USER_SPACE_METHOD uhid
#endif

#define VIRTUAL_DEVICE_METHOD USER_SPACE_METHOD
#define VIRTUAL_DEVICE_CONCAT3(prefix, separator, suffix) prefix ## separator ## suffix
#define VIRTUAL_DEVICE_CONCAT2(prefix, suffix) VIRTUAL_DEVICE_CONCAT3(prefix, _, suffix)
#define VIRTUAL_DEVICE_FUNCTION(function, ...) VIRTUAL_DEVICE_CONCAT2(VIRTUAL_DEVICE_METHOD, \
                                                                      function)(__VA_ARGS__)

// Macros redirecting to specific uinput / uhid implementation
#define pointer_init()              VIRTUAL_DEVICE_FUNCTION(init)
#define pointer_move(motion)        VIRTUAL_DEVICE_FUNCTION(move, motion)
#define pointer_event(kernel)       VIRTUAL_DEVICE_FUNCTION(event, kernel)
#define pointer_destroy(descriptor) VIRTUAL_DEVICE_FUNCTION(destroy, descriptor)

/* Copied from frameworks/native/libs/input/VelocityTracker.cpp ---

// Nanoseconds per milliseconds.
static const nsecs_t NANOS_PER_MS = 1000000;

// Threshold for determining that a pointer has stopped moving.
// Some input devices do not send ACTION_MOVE events in the case where a pointer has
// stopped.  We need to detect this case so that we can accurately predict the
// velocity after the pointer starts moving again.
static const nsecs_t ASSUME_POINTER_STOPPED_TIME = 40 * NANOS_PER_MS;

--- End of copy */

// NOTE: In practice, more than 100 us delay is ensured, so sleeping for 40 ms is enough
#define CURSOR_REFRESH_RATE (40 * 1000000L) // 40 ms in units of [ns], i.e., mouse updated at 25 fps

static struct timespec interval = {
    .tv_sec = 0,
    .tv_nsec = CURSOR_REFRESH_RATE,
};

static struct event_motion motion = {
    .dx = 0,
    .dy = 0,
    .cp = PTHREAD_COND_INITIALIZER,
    .cc = PTHREAD_COND_INITIALIZER,
    .mt = PTHREAD_MUTEX_INITIALIZER,
};

static struct event_kernel kernel = {
    .mt = PTHREAD_MUTEX_INITIALIZER,
};

static pthread_t thread_looper;
static pthread_t thread_poller;

static void * pointer_loop(void * arg) {
    (void) arg;

    int ret;
    struct timespec timeout;

    pthread_mutex_lock(&motion.mt);

    motion.looping = true;

    while (motion.looping) {
        if (motion.dx != 0 || motion.dy != 0) {
            ret = pointer_move(&motion);
            if (ret < 0) {
                LOG_ERRNO("Could not inject motion event");
            } else {
                motion.dx = 0;
                motion.dy = 0;
                pthread_cond_signal(&motion.cp);
            }

            /*ret = clock_gettime(CLOCK_REALTIME, &timeout);
            if (ret < 0) {
                LOG_ERRNO("Could not retrieve current time");

                pthread_mutex_unlock(&motion.mt);
                nanosleep(&interval, NULL);
                pthread_mutex_lock(&motion.mt);
            } else {
                timeout.tv_sec  += interval.tv_sec;
                timeout.tv_nsec += interval.tv_nsec;

                if (timeout.tv_nsec >= 1000000000L) {
                    timeout.tv_nsec -= 1000000000L;
                    timeout.tv_sec++;
                }

                pthread_cond_timedwait(&motion.cp, &motion.mt, &timeout);
            }*/
        } else {
            pthread_cond_wait(&motion.cc, &motion.mt);
        }
    }

    motion.looping = false;

    pthread_mutex_unlock(&motion.mt);

    return NULL;
}

static void * pointer_poll(void * arg) {
    (void) arg;

    int ret;
    struct pollfd pfds[2] = {
        {
            .fd = kernel.fd,
            .events = POLLIN,
            .revents = 0,
        },
        {
            .fd = kernel.efd,
            .events = POLLIN,
            .revents = 0,
        },
    };

    pthread_mutex_lock(&kernel.mt);

    kernel.polling = true;

    while (kernel.polling) {
        pthread_mutex_unlock(&kernel.mt);
        ret = poll(pfds, 2, -1);
        pthread_mutex_lock(&kernel.mt);

        if (ret == 0) {
            continue;
        }

        if (ret < 0) {
            LOG_ERRNO("Could not poll on virtual pointer");
            break;
        }

        if (pfds[0].revents & POLLHUP) {
            LOG_ERROR("Received HUP on virtual pointer");
            break;
        }

        if (pfds[0].revents & POLLIN) {
            LOG_INFO("Received IN on virtual pointer");

            ret = pointer_event(&kernel);
            if (ret < 0) {
                LOG_ERRNO("Could not handle kernel event");
            }
        }
    }

    kernel.polling = false;

    pthread_mutex_unlock(&kernel.mt);

    return NULL;
}

JNIEXPORT jboolean JNICALL
Java_com_genymobile_scrcpy_VirtualPointer_nativeInit(
    JNIEnv * env,
    jclass clazz
) {
    (void) env;
    (void) clazz;

    int fd;
    int efd;
    int ret;

    // Initialize the virtual input device
    fd = pointer_init();
    if (fd < 0) {
        LOG_ERRNO("Could not create virtual pointer");
        return JNI_FALSE;
    }

    // Store the file descriptor
    motion.fd = fd;
    kernel.fd = fd;

    // Prime the kernel event notification
    efd = signal_prime();
    if (efd < 0) {
        LOG_ERRNO("Could not create kernel signal");

        // Remove the virtual pointer
        ret = pointer_destroy(fd);
        if (ret < 0) {
            LOG_ERRNO("Could not destroy virtual pointer");
        }

        return JNI_FALSE;
    }

    // Store the event file descriptor
    kernel.efd = efd;

    // Launch the looper thread
    ret = pthread_create(&thread_looper, NULL, pointer_loop, NULL);
    if (ret != 0) {
        errno = ret;
        LOG_ERRNO("Could not create looper thread");

        // Cancel the kernel signal
        ret = signal_defuse(efd);
        if (ret < 0) {
            LOG_ERRNO("Could not defuse kernel signal");
        }

        // Remove the virtual pointer
        ret = pointer_destroy(fd);
        if (ret < 0) {
            LOG_ERRNO("Could not destroy virtual pointer");
        }

        return JNI_FALSE;
    }

    // Launch the poller thread
    ret = pthread_create(&thread_poller, NULL, pointer_poll, NULL);
    if (ret != 0) {
        errno = ret;
        LOG_ERRNO("Could not create poller thread");

        // Kill the looper thread
        pthread_mutex_lock(&motion.mt);
        motion.looping = false;
        pthread_cond_signal(&motion.cp);
        pthread_cond_signal(&motion.cc);
        pthread_mutex_unlock(&motion.mt);
        pthread_join(thread_looper, NULL);

        // Cancel the kernel signal
        ret = signal_defuse(efd);
        if (ret < 0) {
            LOG_ERRNO("Could not defuse kernel signal");
        }

        // Remove the virtual pointer
        ret = pointer_destroy(fd);
        if (ret < 0) {
            LOG_ERRNO("Could not destroy virtual pointer");
        }

        return JNI_FALSE;
    }

    return JNI_TRUE;
}

JNIEXPORT jboolean JNICALL
Java_com_genymobile_scrcpy_VirtualPointer_nativeKill(
    JNIEnv * env,
    jclass clazz
) {
    (void) env;
    (void) clazz;

    int fd;
    int efd;
    int ret;

    fd = motion.fd;
    efd = kernel.efd;

    // Signal the looper thread that it must stop looping
    pthread_mutex_lock(&motion.mt);
    motion.looping = false;
    pthread_cond_signal(&motion.cp);
    pthread_cond_signal(&motion.cc);
    pthread_mutex_unlock(&motion.mt);

    // Signal the poller thread that it must stop polling
    pthread_mutex_lock(&kernel.mt);
    kernel.polling = false;
    signal_fire(efd);
    pthread_mutex_unlock(&kernel.mt);

    // Join the background threads with the main thread
    pthread_join(thread_looper, NULL);
    pthread_join(thread_poller, NULL);

    // Defuse the kernel event notification
    ret = signal_defuse(efd);
    if (ret < 0) {
        LOG_ERRNO("Could not defuse kernel signal");
        efd = ret;
    }

    // Destroy the virtual input device
    ret = pointer_destroy(fd);
    if (ret < 0) {
        LOG_ERRNO("Could not destroy virtual pointer");
        return JNI_FALSE;
    }

    if (efd < 0) {
        return JNI_FALSE;
    }

    return JNI_TRUE;
}

JNIEXPORT jboolean JNICALL
Java_com_genymobile_scrcpy_VirtualPointer_nativeMove(
    JNIEnv * env,
    jclass clazz,
    jint dx,
    jint dy,
    jboolean calibrating
) {
    (void) env;
    (void) clazz;

    pthread_mutex_lock(&motion.mt);

    // Wait for the looper thread to consume previously produced events when calibrating
    while (calibrating && (motion.dx != 0 || motion.dy != 0)) {
        pthread_cond_wait(&motion.cp, &motion.mt);
    }

    // Signal the looper thread that a new motion event has been received
    motion.dx += dx;
    motion.dy += dy;
    pthread_cond_signal(&motion.cc);

    // Wait for the looper thread to handle the event if the latter is injected in calibration mode
    if (calibrating) {
        pthread_cond_wait(&motion.cp, &motion.mt);
        nanosleep(&interval, NULL);
    }

    pthread_mutex_unlock(&motion.mt);

    return JNI_TRUE;
}
