#ifndef __UHID_H__
#define __UHID_H__

#include "device.h"
#include "event.h"
#include "log.h"

#include <linux/uhid.h>

#include <errno.h>
#include <fcntl.h>
#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>

#define UHID_PATH_NAME "/dev/uhid"

#define UHID_DEVICE_BUS     DEVICE_BUS
#define UHID_DEVICE_NAME    DEVICE_NAME
#define UHID_DEVICE_VENDOR  DEVICE_VENDOR
#define UHID_DEVICE_PRODUCT DEVICE_PRODUCT
#define UHID_DEVICE_VERSION DEVICE_VERSION
#define UHID_DEVICE_COUNTRY DEVICE_COUNTRY

#define UHID_MAX_NAME_SIZE 128

#define UHID_REPORT_ID_MOUSE 0x01

#define UHID_MOUSE_BUTTON_RELEASED 0
#define UHID_MOUSE_BUTTON_PRESSED  1

enum uhid_device_state {
    UHID_DEVICE_STARTED,
    UHID_DEVICE_STOPPED,
    UHID_DEVICE_OPENED,
    UHID_DEVICE_CLOSED,
};

// HID device state
static unsigned char uhid_state = UHID_DEVICE_STOPPED;

// HID device flags
static unsigned char uhid_flags = UHID_DEV_NUMBERED_INPUT_REPORTS;

// HID device descriptor
static unsigned char uhid_rdesc[] = {
    0x05, 0x01,                   /* USAGE_PAGE (Generic Desktop)                 */
    0x09, 0x02,                   /* USAGE (Mouse)                                */
    0xA1, 0x01,                   /* COLLECTION (Application)                     */
    0x09, 0x01,                   /*     USAGE (Pointer)                          */
    0xA1, 0x00,                   /*     COLLECTION (Physical)                    */
#if UHID_REPORT_ID_MOUSE
    0x85, UHID_REPORT_ID_MOUSE,   /*         REPORT_ID (Mouse)                    */
#endif
    0x05, 0x09,                   /*         USAGE_PAGE (Button)                  */
    0x19, 0x01,                   /*             USAGE_MINIMUM (Button L)         */
    0x29, 0x01,                   /*             USAGE_MAXIMUM (Button L)         */
    0x15, 0x00,                   /*             LOGICAL_MINIMUM (0)              */
    0x25, 0x01,                   /*             LOGICAL_MAXIMUM (1)              */
    0x75, 0x01,                   /*             REPORT_SIZE (1)                  */
    0x95, 0x01,                   /*             REPORT_COUNT (1)                 */
    0x81, 0x02,                   /*             INPUT (Data,Var,Abs)             */
    0x75, 0x07,                   /*             REPORT_SIZE (7)                  */
    0x95, 0x01,                   /*             REPORT_COUNT (1)                 */
    0x81, 0x01,                   /*             INPUT (Cnst,Var,Abs)             */
    0x05, 0x01,                   /*         USAGE_PAGE (Generic Desktop)         */
    0x09, 0x30,                   /*             USAGE (X)                        */
    0x09, 0x31,                   /*             USAGE (Y)                        */
    0x17, 0x00, 0x00, 0x00, 0x80, /*             LOGICAL_MINIMUM (-2,147,483,648) */
    0x27, 0xFF, 0xFF, 0xFF, 0x7F, /*             LOGICAL_MAXIMUM (+2,147,483,647) */
    0x75, 0x20,                   /*             REPORT_SIZE (32)                 */
    0x95, 0x02,                   /*             REPORT_COUNT (2)                 */
    0x81, 0x06,                   /*             INPUT (Data,Var,Rel)             */
    0xC0,                         /*     END_COLLECTION                           */
    0xC0,                         /* END_COLLECTION                               */
};

// Interface methods
static int uhid_init();
static int uhid_move(const struct event_motion * const motion);
static int uhid_event(const struct event_kernel * const kernel);
static int uhid_destroy(const int fd);

static int uhid_open(const char * const path, const int flags) {
    int ret;

    ret = open(path, flags);
    if (ret < 0) {
        LOG_ERRNO("Cannot open uhid device");
        return -errno;
    } else {
        return ret;
    }
}

static int uhid_close(const int fd) {
    int ret;

    ret = close(fd);
    if (ret < 0) {
        LOG_ERRNO("Cannot close uhid device");
        return -errno;
    } else {
        return ret;
    }
}

static int uhid_read(const int fd, void * const buf, const size_t num) {
    int ret;

    ret = read(fd, buf, num);
    if (ret == 0) {
        LOG_ERROR("Read EOF from uhid device");
        errno = EINTR;
        return -errno;
    } else if (ret < 0) {
        LOG_ERRNO("Cannot read from uhid device");
        return -errno;
    } else if (ret != num) {
        LOG_ERROR("Wrong size read from uhid device: %d != %zu", ret, num);
        errno = EFAULT;
        return -errno;
    } else {
        return ret;
    }
}

static int uhid_write(const int fd, const void * const buf, const size_t num) {
    int ret;

    ret = write(fd, buf, num);
    if (ret < 0) {
        LOG_ERRNO("Cannot write to uhid device");
        return -errno;
    } else if (ret != num) {
        LOG_ERROR("Wrong size written to uhid device: %d != %zu", ret, num);
        errno = EFAULT;
        return -errno;
    } else {
        return ret;
    }
}

static int uhid_input(const int fd, const int32_t dx, const int32_t dy) {
    struct uhid_event ev;
    unsigned short i;

    i = 0;

    memset(&ev, 0, sizeof(ev));

    ev.type = UHID_INPUT2;

    if (uhid_flags & UHID_DEV_NUMBERED_INPUT_REPORTS) {
        // Report ID (1 byte)
        ev.u.input2.data[i++] = UHID_REPORT_ID_MOUSE;
    }

    // Button L (1 + 7 bits)
    ev.u.input2.data[i++] = UHID_MOUSE_BUTTON_RELEASED;

    // Mouse X & Y (2 x 32 bits)
    ev.u.input2.data[i++] = (dx & 0x000000FFU) >>  0;
    ev.u.input2.data[i++] = (dx & 0x0000FF00U) >>  8;
    ev.u.input2.data[i++] = (dx & 0x00FF0000U) >> 16;
    ev.u.input2.data[i++] = (dx & 0xFF000000U) >> 24;
    ev.u.input2.data[i++] = (dy & 0x000000FFU) >>  0;
    ev.u.input2.data[i++] = (dy & 0x0000FF00U) >>  8;
    ev.u.input2.data[i++] = (dy & 0x00FF0000U) >> 16;
    ev.u.input2.data[i++] = (dy & 0xFF000000U) >> 24;

    // Report size
    ev.u.input2.size = i;

    CHECK(uhid_write(fd, &ev, sizeof(ev)), LOG_ERROR_WRITE);

    return 0;
}

static int uhid_create(const int fd) {
    struct uhid_event ev;

    memset(&ev, 0, sizeof(ev));

    ev.type = UHID_CREATE2;

    ev.u.create2.bus     = UHID_DEVICE_BUS;
    ev.u.create2.vendor  = UHID_DEVICE_VENDOR;
    ev.u.create2.product = UHID_DEVICE_PRODUCT;
    ev.u.create2.version = UHID_DEVICE_VERSION;
    ev.u.create2.country = UHID_DEVICE_COUNTRY;
    ev.u.create2.rd_size = sizeof(uhid_rdesc);
    memcpy(ev.u.create2.rd_data, uhid_rdesc, sizeof(uhid_rdesc));
    strncpy((char *) ev.u.create2.name, UHID_DEVICE_NAME, UHID_MAX_NAME_SIZE);

    CHECK(uhid_write(fd, &ev, sizeof(ev)), LOG_ERROR_WRITE);

    return 0;
}

static int uhid_init() {
    int fd;
    int ret;

    ASSIGN(fd, uhid_open(UHID_PATH_NAME, O_RDWR | O_CLOEXEC | O_NONBLOCK), LOG_ERROR_OPEN);

    ret = uhid_create(fd);
    if (ret < 0) {
        LOG_ERRNO("Could not create uhid device");

        CHECK(uhid_close(fd), LOG_ERROR_CLOSE);

        return ret;
    }

    return fd;
}

static int uhid_move(const struct event_motion * const motion) {
    if (uhid_state != UHID_DEVICE_OPENED) {
        LOG_ERROR("Not ready for sending input report to uhid device");
        errno = EAGAIN;
        return -errno;
    }

    // Inject relative motion as input event
    CHECK(uhid_input(motion->fd, motion->dx, motion->dy), LOG_ERROR_INPUT);

    return 0;
}

static int uhid_event(const struct event_kernel * const kernel) {
    struct uhid_event ev;
    struct uhid_event re;

    memset(&ev, 0, sizeof(ev));

    CHECK(uhid_read(kernel->fd, &ev, sizeof(ev)), LOG_ERROR_READ);

    // Reference: https://www.kernel.org/doc/Documentation/hid/uhid.txt
    switch (ev.type) {
        case UHID_START:
            LOG_INFO("UHID_START from uhid device");
            uhid_state = UHID_DEVICE_STARTED;
            uhid_flags = ev.u.start.dev_flags;
            break;
        case UHID_STOP:
            LOG_INFO("UHID_STOP from uhid device");
            uhid_state = UHID_DEVICE_STOPPED;
            break;
        case UHID_OPEN:
            LOG_INFO("UHID_OPEN from uhid device");
            uhid_state = UHID_DEVICE_OPENED;
            break;
        case UHID_CLOSE:
            LOG_INFO("UHID_CLOSE from uhid device");
            uhid_state = UHID_DEVICE_CLOSED;
            break;
        case UHID_OUTPUT:
            LOG_INFO("UHID_OUTPUT from uhid device");
            switch (ev.u.output.rtype) {
                default:
                    LOG_WARN("Invalid report type: %u", ev.u.output.rtype);
            }
            break;
        case UHID_GET_REPORT:
            LOG_INFO("UHID_GET_REPORT from uhid device");
            memset(&re, 0, sizeof(re));
            re.type = UHID_GET_REPORT_REPLY;
            re.u.get_report_reply.id = ev.u.get_report.id;
            switch (ev.u.get_report.rnum) {
                case UHID_REPORT_ID_MOUSE:
                    LOG_INFO("Report number: UHID_REPORT_ID_MOUSE");
                    switch (ev.u.get_report.rtype) {
                        default:
                            LOG_WARN("Invalid report type: %u", ev.u.get_report.rtype);
                            re.u.get_report_reply.err = EIO;
                    }
                    break;
                default:
                    LOG_WARN("Invalid report number: %u", ev.u.get_report.rnum);
                    re.u.get_report_reply.err = EIO;
            }
            CHECK(uhid_write(kernel->fd, &re, sizeof(re)),
                  "Could not reply to UHID_GET_REPORT for uhid device\n");
            break;
        case UHID_SET_REPORT:
            LOG_INFO("UHID_SET_REPORT from uhid device");
            memset(&re, 0, sizeof(re));
            re.type = UHID_SET_REPORT_REPLY;
            re.u.set_report_reply.id = ev.u.set_report.id;
            switch (ev.u.set_report.rnum) {
                case UHID_REPORT_ID_MOUSE:
                    LOG_INFO("Report number: UHID_REPORT_ID_MOUSE");
                    switch (ev.u.set_report.rtype) {
                        default:
                            LOG_WARN("Invalid report type: %u", ev.u.set_report.rtype);
                            re.u.set_report_reply.err = EIO;
                    }
                    break;
                default:
                    LOG_WARN("Invalid report number: %u", ev.u.set_report.rnum);
                    re.u.set_report_reply.err = EIO;
            }
            CHECK(uhid_write(kernel->fd, &re, sizeof(re)),
                  "Could not reply to UHID_SET_REPORT for uhid device\n");
            break;
        default:
            LOG_WARN("Invalid event from uhid device: %u", ev.type);
    }

    return 0;
}

static int uhid_destroy(const int fd) {
    struct uhid_event ev;

    memset(&ev, 0, sizeof(ev));

    ev.type = UHID_DESTROY;

    CHECK(uhid_write(fd, &ev, sizeof(ev)), LOG_ERROR_WRITE);
    CHECK(uhid_close(fd), LOG_ERROR_CLOSE);

    return 0;
}

#endif /* __UHID_H__ */
