#ifndef __UINPUT_H__
#define __UINPUT_H__

#include "device.h"
#include "event.h"
#include "log.h"

#include <linux/uinput.h>

#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>

#define UINPUT_PATH_NAME "/dev/uinput"

#define UINPUT_DEVICE_BUS     DEVICE_BUS
#define UINPUT_DEVICE_NAME    DEVICE_NAME
#define UINPUT_DEVICE_VENDOR  DEVICE_VENDOR
#define UINPUT_DEVICE_PRODUCT DEVICE_PRODUCT
#define UINPUT_DEVICE_VERSION DEVICE_VERSION

#define SYN_VALUE 0

// Interface methods
static int uinput_init();
static int uinput_move(const struct event_motion * const motion);
static int uinput_event(const struct event_kernel * const kernel);
static int uinput_destroy(const int fd);

static int uinput_open(const char * const path, const int flags) {
    int ret;

    ret = open(path, flags);
    if (ret < 0) {
        LOG_ERRNO("Cannot open uinput device");
        return -errno;
    } else {
        return ret;
    }
}

static int uinput_close(const int fd) {
    int ret;

    ret = close(fd);
    if (ret < 0) {
        LOG_ERRNO("Cannot close uinput device");
        return -errno;
    } else {
        return ret;
    }
}

static int uinput_read(const int fd, void * const buf, const size_t num) {
    int ret;

    ret = read(fd, buf, num);
    if (ret == 0) {
        LOG_ERROR("Read EOF from uinput device");
        errno = EINTR;
        return -errno;
    } else if (ret < 0) {
        LOG_ERRNO("Cannot read from uinput device");
        return -errno;
    } else if (ret != num) {
        LOG_ERROR("Wrong size read from uinput device: %d != %zu", ret, num);
        errno = EFAULT;
        return -errno;
    } else {
        return ret;
    }
}

static int uinput_write(const int fd, const void * const buf, const size_t num) {
    int ret;

    ret = write(fd, buf, num);
    if (ret < 0) {
        LOG_ERRNO("Cannot write to uinput device");
        return -errno;
    } else if (ret != num) {
        LOG_ERROR("Wrong size written to uinput device: %d != %zu", ret, num);
        errno = EFAULT;
        return -errno;
    } else {
        return ret;
    }
}

static int uinput_ioctl(const int fd, const unsigned long req, ...) {
    int ret;
    va_list valist;
    const void * arg;

    va_start(valist, req);
    arg = va_arg(valist, const void *);

    ret = ioctl(fd, req, arg);
    if (ret < 0) {
        LOG_ERRNO("Cannot ioctl uinput device");
        return -errno;
    } else {
        return ret;
    }
}

static int uinput_input(const int fd, const __u16 type, const __u16 code, const __s32 value) {
    struct input_event ie;

    ie.type = type;
    ie.code = code;
    ie.value = value;
    // Timestamp values below are ignored
    // (events are timestamped by the evdev kernel module)
    ie.time.tv_sec = 0;
    ie.time.tv_usec = 0;

    CHECK(uinput_write(fd, &ie, sizeof(ie)), LOG_ERROR_WRITE);

    return 0;
}

static int uinput_create(const int fd) {
    struct uinput_config {
        struct input_id id;
        char name[UINPUT_MAX_NAME_SIZE];
        __u32 ff_effects_max;
        __s32 absval [ABS_CNT];
        __s32 absres [ABS_CNT];
        __s32 absmin [ABS_CNT];
        __s32 absmax [ABS_CNT];
        __s32 absfuzz[ABS_CNT];
        __s32 absflat[ABS_CNT];
    } uconfig;
    struct uinput_setup usetup;
    struct uinput_user_dev uud;
    int version;
    int rc;

    ASSIGN(rc, uinput_ioctl(fd, UI_GET_VERSION, &version), LOG_ERROR_IOCTL);

    CHECK(uinput_ioctl(fd, UI_SET_EVBIT, EV_KEY), LOG_ERROR_IOCTL);
    CHECK(uinput_ioctl(fd, UI_SET_KEYBIT, BTN_MOUSE), LOG_ERROR_IOCTL);

    CHECK(uinput_ioctl(fd, UI_SET_EVBIT, EV_REL), LOG_ERROR_IOCTL);
    CHECK(uinput_ioctl(fd, UI_SET_RELBIT, REL_X), LOG_ERROR_IOCTL);
    CHECK(uinput_ioctl(fd, UI_SET_RELBIT, REL_Y), LOG_ERROR_IOCTL);

    memset(&uconfig, 0, sizeof(uconfig));

    uconfig.id.bustype  = UINPUT_DEVICE_BUS;
    uconfig.id.vendor   = UINPUT_DEVICE_VENDOR;
    uconfig.id.product  = UINPUT_DEVICE_PRODUCT;
    uconfig.id.version  = UINPUT_DEVICE_VERSION;
    strncpy(uconfig.name, UINPUT_DEVICE_NAME, UINPUT_MAX_NAME_SIZE);

    if (rc == 0 && version >= 5) {
        memset(&usetup, 0, sizeof(usetup));
        memcpy(&usetup.id,             &uconfig.id,             sizeof(uconfig.id));
        memcpy(&usetup.name,           &uconfig.name,           sizeof(uconfig.name));
        memcpy(&usetup.ff_effects_max, &uconfig.ff_effects_max, sizeof(uconfig.ff_effects_max));
        CHECK(uinput_ioctl(fd, UI_DEV_SETUP, &usetup), LOG_ERROR_IOCTL);
        CHECK(uinput_ioctl(fd, UI_DEV_CREATE), LOG_ERROR_IOCTL);
    } else {
        memset(&uud   , 0, sizeof(uud));
        memcpy(&uud   .id,             &uconfig.id,             sizeof(uconfig.id));
        memcpy(&uud   .name,           &uconfig.name,           sizeof(uconfig.name));
        memcpy(&uud   .ff_effects_max, &uconfig.ff_effects_max, sizeof(uconfig.ff_effects_max));
        memcpy(&uud   .absmax,         &uconfig.absmax,         sizeof(uconfig.absmax));
        memcpy(&uud   .absmin,         &uconfig.absmin,         sizeof(uconfig.absmin));
        memcpy(&uud   .absfuzz,        &uconfig.absfuzz,        sizeof(uconfig.absfuzz));
        memcpy(&uud   .absflat,        &uconfig.absflat,        sizeof(uconfig.absflat));
        CHECK(uinput_write(fd, &uud, sizeof(uud)), LOG_ERROR_WRITE);
        CHECK(uinput_ioctl(fd, UI_DEV_CREATE), LOG_ERROR_IOCTL);
    }

    return 0;
}

static int uinput_init() {
    int fd;
    int ret;

    ASSIGN(fd, uinput_open(UINPUT_PATH_NAME, O_RDWR | O_CLOEXEC | O_NONBLOCK), LOG_ERROR_OPEN);

    ret = uinput_create(fd);
    if (ret < 0) {
        LOG_ERRNO("Could not create uinput device");

        CHECK(uinput_close(fd), LOG_ERROR_CLOSE);

        return ret;
    }

    return fd;
}

static int uinput_move(const struct event_motion * const motion) {
    // Inject relative motion as input event
    CHECK(uinput_input(motion->fd, EV_REL, REL_X, motion->dx), LOG_ERROR_INPUT);
    CHECK(uinput_input(motion->fd, EV_REL, REL_Y, motion->dy), LOG_ERROR_INPUT);
    CHECK(uinput_input(motion->fd, EV_SYN, SYN_REPORT, SYN_VALUE), LOG_ERROR_INPUT);

    return 0;
}

static int uinput_event(const struct event_kernel * const kernel) {
    struct input_event ev;

    memset(&ev, 0, sizeof(ev));

    CHECK(uinput_read(kernel->fd, &ev, sizeof(ev)), LOG_ERROR_READ);

    // Reference: https://www.kernel.org/doc/Documentation/input/event-codes.txt
    switch (ev.type) {
        case EV_KEY:
            LOG_INFO("EV_KEY from uinput device");
            switch (ev.code) {
                case BTN_MOUSE:
                    LOG_INFO("BTN_MOUSE: %d", ev.value);
                    break;
                default:
                    LOG_WARN("Invalid code for EV_KEY: %u", ev.code);
            }
            break;
        case EV_REL:
            LOG_INFO("EV_REL from uinput device");
            switch (ev.code) {
                case REL_X:
                    LOG_INFO("REL_X: %d", ev.value);
                    break;
                case REL_Y:
                    LOG_INFO("REL_Y: %d", ev.value);
                    break;
                default:
                    LOG_WARN("Invalid code for EV_REL: %u", ev.code);
            }
            break;
        case EV_SYN:
            LOG_INFO("EV_SYN from uinput device");
            switch (ev.code) {
                case SYN_REPORT:
                    LOG_INFO("SYN_REPORT: %d", ev.value);
                    break;
                case SYN_DROPPED:
                    LOG_INFO("SYN_DROPPED: %d", ev.value);
                    break;
                default:
                    LOG_WARN("Invalid code for EV_SYN: %u", ev.code);
            }
            break;
        default:
            LOG_WARN("Invalid event from uinput device: %u", ev.type);
    }

    return 0;
}

static int uinput_destroy(const int fd) {
    CHECK(uinput_ioctl(fd, UI_DEV_DESTROY), LOG_ERROR_IOCTL);
    CHECK(uinput_close(fd), LOG_ERROR_CLOSE);

    return 0;
}

#endif /* __UINPUT_H__ */
